#include <iostream.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <AudioDevices/OSSAudioDevice.h>

const int OSSAudioDevice::MAX_NUM_DEVS = 4;
const int OSSAudioDevice::MAX_NUM_MIXERS = 4;
const int OSSAudioDevice::MAX_OSS_NAME_SIZE = 64;
const int OSSAudioDevice::MAX_NUM_OSS_DEVS = 4;

// stores the probed device names
const string audioDevNames[OSSAudioDevice::MAX_NUM_DEVS]; 
const string AUDIO_STR = "/dev/audio";
const string MIXER_STR = "/dev/mixer";

DeviceCaps*
OSSAudioDevice::getDevCaps()
{
    DeviceCaps *pDeviceCaps = new DeviceCaps();

    for(int dev = 0;dev < MAX_NUM_OSS_DEVS;dev++)
    {
	if(getAudioCaps(dev,pDeviceCaps) != true)
	{
	    // AUDIT Could not get audio device capabilities i.
	}

	if(getMixerCaps(dev,pDeviceCaps) != true)
	{
	    // AUDIT Could not get mixer device capabilities i.
	}
    }
}

// This function collects together information about various devices, so
// it can be presented in a NAL way.
bool 
OSSAudioDevice::getAudioCaps(int audioDevNum, DeviceCaps *deviceCaps)
{
    // Device numbers seems to start at one, but there is still a /dev/audio, so
    // it needs to be probed as well. The assumption is that the Linux
    // distibution is not too old, so /dev will follow the regular Linux standards
    // for device naming.

    // the file system /dev name
    string deviceName; 
    ostream devStr;
    if(audioDevNum == 0)
    {
	devStr << AUDIO_STR << ends;
	deviceName = devStr.str();
    }
    else
    {
	devStr << AUDIO_STR << audioDevNum << ends;
	deviceName = devStr.str();
    }

    int audioDesc;
    audioDesc = open(deviceName, O_RDWR);
    audioDevNames[audioDevNum] = deviceName;

    if(audioDesc < 0)
    {
	// AUDIT The audio audioDevNum was not found.
	return FALSE;
    }

    // Checking if the device is full duplex must be done first. Only full
    // duplex decives will be supported for now.
    if(ioctl(audioDesc, SNDCTL_DSP_SETDUPLEX, 0) == -1)
    {
	// AUDIT Only full duplex devices are supported
	//deviceCaps->setIsDuplex(audioDevNum, false);
	return FALSE;
    }
    else
    {
	//deviceCaps->setIsDuplex(audioDevNum, true);
    }

    AudioFormats *audioFormats = new AudioFormats();

    // Check which audio encodings are supported. This only reports what the
    // hardware acutually supports. No driver level sofware encodings are
    // reported. This is probably not a bad thing because we want decent
    // performance. Prior to version 3.6 of OSS MU_LAW is always reported as
    // being supported.
    int mask;
    if(ioctl(audioDesc, SNDCTL_DSP_GETFMTS, &mask) == -1)
    {
	// AUDIT Could not get encoding mask.
    }
    else
    {
	for(int encoding = 0;encoding < NUM_ENCODINGS;encoding++)
	{
	    if(mask & NALEncoding2OSS(encoding))
	    {
		audioFormats->setEncoding(sampleRates[encoding]);
	    }
	}
    }

    // Check which sampling modes are supported.
    for(int rate = 0;rate < NUM_SAMPLE_RATES;rate++)
    {
	int sampleRate = rate;
	if(ioctl(audioDesc, SNDCTL_DSP_SPEED, &sampleRate) == -1)
	{
	    // AUDIT Sample rate AudioFormat::sampleRates[rate] is not supported.
	}
	else
	{
	    int chan = AudioFormat::MONO;
	    if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &chan) == -1)
	    {
		// AUDIT MONO not supported.
	    }
	    else
	    {
		AutoFormats audioFormats = new AutoFormats();	
		audioFormat->sampleRate = AudioFormat::sampleRates[rate];
		audioFormat->numChannels = AudioFormat::MONO;
		audioFormats->setAudioFormat(rate, audioFormat);
	    }
	    chan = AudioFormat::STEREO;
	    if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &chan) == -1)
	    {
		// AUDIT STEREO not supported.
	    }
	    else
	    {
		AudioFormat = new AudioFormat();	
		audioFormat->sampleRate = AudioFormat::sampleRates[rate];
		audioFormat->numChannels = AudioFormat::STEREO;
		audioFormats->setAudioFormat(rate, audioFormat);
	    }
	}
    }

    deviceCaps->setAudioFormat(audioDevNum, audioFormat);
    close(audioDesc);
    return TRUE;
}

int
OSSAudioDevice::NALEncoding2OSS(AudioFormat::Encoding encoding)
{
    switch(encoding)
    {
	case AudioFormat::PCMU: return AFMT_MU_LAW;
	case AudioFormat::PCMA: return AFMT_A_LAW;
	case AudioFormat::S8:   return AFMT_S8;
	case AudioFormat::U8:   return AFMT_U8;
	case AudioFormat::S16:  return AFMT_S16_NE;
	default:
    }

    return AudioFormat::NO_FORMAT;
}

// The mixers are only probed to make sure some needed capabilities exist. Only
// the name of the device is saved (the name on the filesystem).
bool
OSSAudioDevice::getMixerCaps(int mixerNum, DeviceCaps *deviceCaps)
{
    // Device numbers seems to start at one, but there is still a /dev/mixer, so
    // it needs to be probed as well. The assumption is that the Linux
    // distibution is not too old, so /dev will follow the regular Linux standards
    // for device naming.
    string deviceName; 
    ostream devStr;
    if(mixerNum == 0)
    {
	devStr << MIXER_STR << ends;
	deviceName = devStr.str();
    }
    else
    {
	devStr << MIXER_STR << mixerNum << ends;
	deviceName = devStr.str();
    }

    int mixerDesc;
    mixerDesc = open(deviceName, O_RDWR);

    if(mixerDesc < 0)
    {
	// AUDIT The mixer mixerNum was not found.
	return FALSE;
    }

    mixer_info mixerInfo;
    ostream ossName;
    if(ioctl(mixerDesc, SOUND_MIXER_INFO, &mixerInfo) == -1)
    {
	// AUDIT Could not get mixer info.
    }
    else
    {
	ossName << mixerInfo.name << ends;
	// should the name really be the mixer name??
	deviceCaps->setName(ossName.str());
    }

    int mask;
    if(ioctl(mixerDesc, SOUND_MIXER_READ_DEVMASK, &mask) == -1)
    {
	if(mask & SOUND_MIXER_PCM)
	{
	    deviceCaps->setInVolSupp(mixerNum, true);
	}
	else
	{
	    deviceCaps->setInVolSupp(mixerNum, false);
	}

	if(mask & SOUND_MIXER_MIC)
	{
	    deviceCaps->setInVolSupp(mixerNum, true);

	}
	else
	{
	    deviceCaps->setInVolSupp(mixerNum, false);
	}
    }
    else
    {
	// AUDIT Could not get device mask from mixer.
    }

    close(mixerDesc);
    return TRUE;
}

int
OSSAudioDevice::open()
{
    audioDesc = open(audioDevName, O_RDWR | O_NDELAY);
    if(audioDesc == 0)
    {
	// AUDIT Could not open device.
	return FALSE;
    }
   
    // This has to be done first.
    if(ioctl(audioDesc, SNDCTL_DSP_SETDUPLEX, 0) == -1)
    {
	// AUDIT Could not set device to duplex.
	close();
	return FALSE;
    }
 
    // Sampling parameters must be set before the blocksize can be queried.
    int format = NALEncoding2OSS(audioFormats->getEncodingToUse(devCap->devHandle));
    if(ioctl(audioDesc, SNDCTL_DSP_SETFMT, &format) == -1)
    {
	// AUDIT The device does not support requested format.
    }

    int rate = devCap->audioFormat->getSampleRateToUse(devCap->devHandle);
    if(ioctl(audioDesc, SNDCTL_DSP_SPEED, &rate) == -1)
    {
	// AUDIT The device does not support requested sampling rate.
    }

    int channel = devCap->audioFormat->getChannelToUse(devCap->devHandle);
    if(ioctl(audioDesc, SNDCTL_DSP_STEREO, &rate) == -1)
    {
	// AUDIT The device does not support requested sampling rate.
    }

    // The device automatically computes the fragment size (blocksize) after
    // sampling parameters have been set. For now the blocksize will be
    // calculated this way.
    int fragSize;
    if(ioctl(audioDesc,SNDCTL_DSP_GETBLKSIZE, &fragSize) == -1)
    {
	// AUDIT Could not get the fragment size.
    }
    else
    {
	audioFormats->writeBlockSize = readBlockSize = fragSize;
    }
}

bool 
OSSAudioDevice::isOpen()
{
    return deviceOpen;
}

void
OSSAudioDevice::close()
{
    close(audioDesc);
    deviceOpen = false;
}

int
OSSAudioDevice::read(sample *appBuff, int buffSize)
{
    audio_buf_info bufInfo;
    ioctl(audioDesc, SNDCTL_DSP_GETISPACE, &bufInfo);
    int bytesAvaiable = min(bufInfo.bytes, buffSize);
    int bytesRead = read(audioDesc, (char*)appBuff, bytesAvaiable);

    // AUDIT buffSize bytesAvaiable bytesRead

    return bytesRead;
}

int
OSSAudioDevice::write(sample *appBuff, int buffSize)
{
    char *buffPointer = (char*)appBuff;
    int thisWrite = 0, toWrite = buffSize, written = 0;
    for(; written <= buffSize; written += thisWrite, buffPointer += thisWrite,
        toWrite -= thisWrite)
    {
	thisWrite = write(audioDesc,buffPointer,appBytesLeft);
    }

    return written;
}

// ioctl should be done on the /dev/audio device. There is no way to correlate
// mixers and audio devices, so all the volume functions use the audio device
// file descriptor.

// It is assuemd that only microphone line needs to be adjusted.
void
OSSAudioDevice::setInputVol(int vol)
{
    int volume = vol2Dev(vol); 
    if(ioctl(audioDesc,MIXER_WRITE(SOUND_MIXER_MIC),&volume) == -1)
    {
	// AUDIT Could not set input volume.
    }
}

int
OSSAudioDevice::getInputVol()
{
    int volume;
    if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_MIC),&volume) == -1)
    {
	// AUDIT Could not set input volume.
    }

    return volume & 0xff;
}

void
OSSAudioDevice::setOutputVol(int vol)
{
    int volume = vol2Dev(vol); 
    if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_PCM),&volume) == -1)
    {
	// AUDIT Could not set input volume.
    }
}

int
OSSAudioDevice::getOutputVol()
{
    int volume;
    if(ioctl(audioDesc,MIXER_READ(SOUND_MIXER_MIC),&volume) == -1)
    {
	// AUDIT Could not set input volume.
    }

    return volume & 0x000000ff; // Extract the left volume.
}

static int
OSSAudioDevice::vol2Dev(int vol)
{
    int volume = (vol << 8) | vol;
    return volume;
}
