#include "Codec.h"

#include <iostream>

using namespace std;

const int G711_SAMPLES_PER_FRAME = 160;
const int TABLE_SIZE = 8;
const int BIAS = 0x84;		/* Bias for linear code. */
const int CLIP = 8159;
const int SIGN_BIT = 0x80;	/* Sign bit for a A-law byte. */
const int QUANT_MASK = 0xf;  /* Quantization field mask. */
const int NSEGS = 8;         /* Number of A-law segments. */
const int SEG_SHIFT = 4;     /* Left shift for segment number. */
const int SEG_MASK = 0x70;   /* Segment field mask. */

class G711 : public Codec
{
public: 
	virtual void decode(unsigned char *inbuf, short *outbuf, int &length);
	virtual void encode(short *inbuf, unsigned char *outbuf, int &length);

	virtual const CodecFormatList &getSupportedFormats(){
		return formatlist;
	}

	G711();

protected:
	static short searchAEnd(short val){
		return search(val,seg_aend);
	}

	static short searchUEnd(short val){
		return search(val,seg_uend);
	}

	//serach for value in table.
	//returns location or TABLE_SIZE if not found
	static short search(short val,short *table){
	   short i;
   
	   for (i = 0; i < TABLE_SIZE; i++) {
		  if (val <= table[i])
			return (i);
	   }

	   return (TABLE_SIZE);
	}

	/*
	 * linear2alaw() - Convert a 16-bit linear PCM value to 8-bit A-law
	 *
	 * linear2alaw() accepts an 16-bit integer and encodes it as A-law data.
	 *
	 *		Linear Input Code	Compressed Code
	 *	------------------------	---------------
	 *	0000000wxyza			000wxyz
	 *	0000001wxyza			001wxyz
	 *	000001wxyzab			010wxyz
	 *	00001wxyzabc			011wxyz
	 *	0001wxyzabcd			100wxyz
	 *	001wxyzabcde			101wxyz
	 *	01wxyzabcdef			110wxyz
	 *	1wxyzabcdefg			111wxyz
	 *
	 * For further information see John C. Bellamy's Digital Telephony, 1982,
	 * John Wiley & Sons, pps 98-111 and 472-476.
	 */
	static unsigned char linear2alaw(short pcm_val);
	//alaw2linear() - Convert an A-law value to 16-bit linear PCM
	static short alaw2linear(unsigned char	a_val);

	/*
	* linear2ulaw() - Convert a linear PCM value to u-law
	*
	* In order to simplify the encoding process, the original linear magnitude
	* is biased by adding 33 which shifts the encoding range from (0 - 8158) to
	* (33 - 8191). The result can be seen in the following encoding table:
	*
	*	Biased Linear Input Code	Compressed Code
	*	------------------------	---------------
	*	00000001wxyza			000wxyz
	*	0000001wxyzab			001wxyz
	*	000001wxyzabc			010wxyz
	*	00001wxyzabcd			011wxyz
	*	0001wxyzabcde			100wxyz
	*	001wxyzabcdef			101wxyz
	*	01wxyzabcdefg			110wxyz
	*	1wxyzabcdefgh			111wxyz
	*
	* Each biased linear code has a leading 1 which identifies the segment
	* number. The value of the segment number is equal to 7 minus the number
	* of leading 0's. The quantization interval is directly available as the
	* four bits wxyz.  * The trailing bits (a - h) are ignored.
	*
	* Ordinarily the complement of the resulting code word is used for
	* transmission, and so the code word is complemented before it is returned.
	*
	* For further information see John C. Bellamy's Digital Telephony, 1982,
	* John Wiley & Sons, pps 98-111 and 472-476.
	*/

	// 2's complement (16-bit range)
	static unsigned char linear2ulaw(short pcm_val);

	/*
	 * ulaw2linear() - Convert a u-law value to 16-bit linear PCM
	 *
	 * First, a biased linear code is derived from the code word. An unbiased
	 * output can then be obtained by subtracting 33 from the biased code.
	 *
	 * Note that this function expects to be passed the complement of the
	 * original code word. This is in keeping with ISDN conventions.
	 */
	static short ulaw2linear(unsigned char u_val);

private:
	CodecFormatList formatlist;

	virtual void destroy(){
		delete this;
	}

	static unsigned char u2a[128];
	static unsigned char a2u[128];

	static short seg_aend[TABLE_SIZE];
	static short seg_uend[TABLE_SIZE];

	short mulawtolin[256];
	unsigned char lintomulaw[65536];

	short alawtolin[256];
	unsigned char lintoalaw[8192]; 

};

CODEC_FACTORY_ENTRY( G711, g711 )

short G711::seg_aend[TABLE_SIZE] = {0x1F, 0x3F, 0x7F, 0xFF,
			    0x1FF, 0x3FF, 0x7FF, 0xFFF};

short G711::seg_uend[TABLE_SIZE] = {0x3F, 0x7F, 0xFF, 0x1FF,
			    0x3FF, 0x7FF, 0xFFF, 0x1FFF};

/* copy from CCITT G.711 specifications */

/* u- to A-law conversions */
unsigned char G711::u2a[128] = {			
	1,	1,	2,	2,	3,	3,	4,	4,
	5,	5,	6,	6,	7,	7,	8,	8,
	9,	10,	11,	12,	13,	14,	15,	16,
	17,	18,	19,	20,	21,	22,	23,	24,
	25,	27,	29,	31,	33,	34,	35,	36,
	37,	38,	39,	40,	41,	42,	43,	44,
	46,	48,	49,	50,	51,	52,	53,	54,
	55,	56,	57,	58,	59,	60,	61,	62,
	64,	65,	66,	67,	68,	69,	70,	71,
	72,	73,	74,	75,	76,	77,	78,	79,
	80,	82,	83,	84,	85,	86,	87,	88,
	89,	90,	91,	92,	93,	94,	95,	96,
	97, 98, 99, 100, 101, 102, 103, 104,
	105, 106, 107, 108, 109, 110, 111, 112,
	113, 114, 115, 116, 117, 118, 119, 120,
	121, 122, 123, 124, 125, 126, 127, 128
};

/* A- to u-law conversions */
unsigned char G711::a2u[128] = {
	1,	3,	5,	7,	9,	11,	13,	15,
	16,	17,	18,	19,	20,	21,	22,	23,
	24,	25,	26,	27,	28,	29,	30,	31,
	32,	32,	33,	33,	34,	34,	35,	35,
	36,	37,	38,	39,	40,	41,	42,	43,
	44,	45,	46,	47,	48,	48,	49,	49,
	50,	51,	52,	53,	54,	55,	56,	57,
	58,	59,	60,	61,	62,	63,	64,	64,
	65,	66,	67,	68,	69,	70,	71,	72,
	73,	74,	75,	76,	77,	78,	79,	80,
	80,	81,	82,	83,	84,	85,	86,	87,
	88,	89,	90,	91,	92,	93,	94,	95,
	96,	97,	98,	99,	100, 101, 102, 103,
	104, 105, 106, 107, 108, 109, 110, 111,
	112, 113, 114, 115, 116, 117, 118, 119,
	120, 121, 122, 123, 124, 125, 126, 127
};

unsigned char G711::linear2alaw(short pcm_val)
{
   short	 mask;
   short	 seg;
   unsigned char aval;
   
   pcm_val = pcm_val >> 3;

   if (pcm_val >= 0) {
      mask = 0xD5;		/* sign (7th) bit = 1 */
   } else {
      mask = 0x55;		/* sign bit = 0 */
      pcm_val = -pcm_val - 1;
   }
   
   /* Convert the scaled magnitude to segment number. */
   seg = searchAEnd(pcm_val);
   
   /* Combine the sign, segment, and quantization bits. */
   
   if (seg >= 8)		/* out of range, return maximum value. */
      return (unsigned char) (0x7F ^ mask);
   else {
      aval = (unsigned char) seg << SEG_SHIFT;
      if (seg < 2)
	 aval |= (pcm_val >> 1) & QUANT_MASK;
      else
	 aval |= (pcm_val >> seg) & QUANT_MASK;
      return (aval ^ mask);
   }
}

short G711::alaw2linear(unsigned char	a_val)
{
   short t;
   short seg;
   
   a_val ^= 0x55;
   
   t = (a_val & QUANT_MASK) << 4;
   seg = ((unsigned)a_val & SEG_MASK) >> SEG_SHIFT;
   switch (seg) {
   case 0:
      t += 8;
      break;
   case 1:
      t += 0x108;
      break;
   default:
      t += 0x108;
      t <<= seg - 1;
   }
   return ((a_val & SIGN_BIT) ? t : -t);
}

unsigned char G711::linear2ulaw(short pcm_val)
{
   short         mask;
   short	 seg;
   unsigned char uval;
   
   /* Get the sign and the magnitude of the value. */
   pcm_val = pcm_val >> 2;
   if (pcm_val < 0) {
      pcm_val = -pcm_val;
      mask = 0x7F;
   } else {
      mask = 0xFF;
   }
   if ( pcm_val > CLIP ) pcm_val = CLIP;		/* clip the magnitude */
   pcm_val += (BIAS >> 2);
   
   /* Convert the scaled magnitude to segment number. */
   seg = searchUEnd(pcm_val);
   
   /*
   * Combine the sign, segment, quantization bits;
   * and complement the code word.
   */
   if (seg >= 8)		/* out of range, return maximum value. */
      return (unsigned char) (0x7F ^ mask);
   else {
      uval = (unsigned char) (seg << 4) | ((pcm_val >> (seg + 1)) & 0xF);
      return (uval ^ mask);
   }
}

short G711::ulaw2linear(unsigned char u_val)
{
   short t;
   
   /* Complement to obtain normal u-law value. */
   u_val = ~u_val;
   
   /*
    * Extract and bias the quantization bits. Then
    * shift up by the segment number and subtract out the bias.
    */
   t = ((u_val & QUANT_MASK) << 3) + BIAS;
   t <<= ((unsigned)u_val & SEG_MASK) >> SEG_SHIFT;
   
   return ((u_val & SIGN_BIT) ? (BIAS - t) : (t - BIAS));
}

G711::G711()
{
	/* 8kHz -law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-8K-Mono",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			0,0,160,
			AudioFormat(AudioFormat::SIGNED16,8000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-8K-Stereo",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			83,0,320,
			AudioFormat(AudioFormat::SIGNED16,8000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	/* 16kHz -law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-16K-Mono",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			84,0,160,
			AudioFormat(AudioFormat::SIGNED16,16000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-16K-Stereo",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			85,0,320,
			AudioFormat(AudioFormat::SIGNED16,16000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);
	/* 32kHz -law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-32K-Mono",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			86,0,160,
			AudioFormat(AudioFormat::SIGNED16,32000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-32K-Stereo",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			87,0,320,
			AudioFormat(AudioFormat::SIGNED16,32000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);
	/* 48kHz -law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-48K-Mono",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			88,0,160,
			AudioFormat(AudioFormat::SIGNED16,48000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-48K-Stereo",
			"ITU G.711 -law codec.  Sun Microsystems public implementation.",
			89,0,320,
			AudioFormat(AudioFormat::SIGNED16,48000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	/* 8kHz A-law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-8K-Mono",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			8,0,160,
			AudioFormat(AudioFormat::SIGNED16,8000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-8K-Stereo",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			90,0,320,
			AudioFormat(AudioFormat::SIGNED16,8000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	/* 16kHz A-law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-16K-Mono",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			91,0,160,
			AudioFormat(AudioFormat::SIGNED16,16000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-16K-Stereo",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			124,0,320,
			AudioFormat(AudioFormat::SIGNED16,16000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);
	/* 32kHz A-law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-32K-Mono",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			92,0,160,
			AudioFormat(AudioFormat::SIGNED16,32000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-32K-Stereo",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			93,0,320,
			AudioFormat(AudioFormat::SIGNED16,32000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);
	/* 48kHz A-law */
	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-48K-Mono",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			94,0,160,
			AudioFormat(AudioFormat::SIGNED16,48000,16,1,
				G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);

	formatlist.insert(formatlist.end(),
		CodecFormat(
			"PCMU-48K-Stereo",
			"ITU G.711 A-law codec.  Sun Microsystems public implementation.",
			95,0,320,
			AudioFormat(AudioFormat::SIGNED16,48000,16,2,
				2 * G711_SAMPLES_PER_FRAME * AudioFormat::BYTES_PER_SAMPLE
			)
		)
	);


    int i;
    
    for(i = 0; i < 256; i++)
            mulawtolin[i] = ulaw2linear((unsigned char)i);

    for(i = -32767; i < 32768; i++) 
            lintomulaw[(unsigned short)i] = linear2ulaw((short)i);

    for(i = 0; i < 256; i++) 
            alawtolin[i] = alaw2linear((unsigned char)i);

    for(i = -32767; i < 32768; i+= 8) 
            lintoalaw[(unsigned short)i>>3] = linear2alaw((short)i);

	codecState = NULL;
	
}

void G711::encode(short *inbuf, unsigned char *outbuf, int &length)
{
	cout << "G711 encode" << endl;

	int len;
    unsigned char *p, *pe;

//    assert(idx < G711_NUM_FORMATS);
//    UNUSED(state);

    /* No difference between A-law and U-law sizes */
    //len = cs[idx].mean_coded_frame_size;
	len = codecFormat->getFrameSize();
    
	outbuf = new unsigned char[len];

    p  = outbuf;
    pe = outbuf + len;

	if(codecFormat->getName().find("PCMU") != string::npos){
		while(p != pe) {
			*p = lintomulaw[(*inbuf)];
			p++; inbuf++;
		}
    } else {
        while(p != pe) {
            *p = lintoalaw[(*inbuf)];
            p++; inbuf++;
        }
    }
    
    length = len;
}

void G711::decode(unsigned char *inbuf, short *outbuf, int &length)
{
	cout << "G711 decode" << endl;

    int len;
    unsigned char *p, *pe;

	len = codecFormat->getFrameSize();

    p  = inbuf;
    pe = inbuf + len;

	if(codecFormat->getName().find("PCMU") != string::npos){
        while(p != pe) {
            *outbuf = mulawtolin[(*p)];
            outbuf++; p++;
        }
    } else {
        while(p != pe) {
            *outbuf = alawtolin[(*p)];
            outbuf++; p++;
        }
    }

    length = len;
}